library ieee;
use ieee.std_logic_1164.all;
use work.vme64x_pkg.all;

package vme64x_sim_pkg is
  constant c_log : boolean := false;

  constant slave_ga : std_logic_vector (4 downto 0) := b"0_1101";

  subtype cfg_addr_t is std_logic_vector (19 downto 0);
  subtype byte_t is std_logic_vector (7 downto 0);
  subtype word_t is std_logic_vector (15 downto 0);
  subtype lword_t is std_logic_vector (31 downto 0);
  subtype qword_t is std_logic_vector (63 downto 0);

  type word_array_t is array (natural range <>) of word_t;
  type lword_array_t is array (natural range <>) of lword_t;
  type qword_array_t is array (natural range <>) of qword_t;

  subtype vme_am_t is std_logic_vector (5 downto 0);
  subtype vme_xam_t is std_logic_vector (7 downto 0);

  --  Debug utility.
  function hex (v : std_logic_vector) return string;

  --  Bus timer.
  --  Set TIMEOUT to true when the slave has not ack-ed the transaction.
  procedure bus_timer (signal vme_in : in  t_vme64x_in;
                       signal timeout : out boolean;
                       signal clk : std_logic);

  --  Write a byte to the CS/CSR space.
  procedure write8_conf (signal vme_in : out t_vme64x_in;
                         signal vme_out : in t_vme64x_out;
                         signal vme_timeout : boolean;
                         addr : cfg_addr_t;
                         data : byte_t);

  --  Read a byte from the CS/CSR space.
  procedure read8_conf (signal vme_in : out t_vme64x_in;
                        signal vme_out : in t_vme64x_out;
                        signal vme_timeout : boolean;
                        addr : cfg_addr_t;
                        variable data : out byte_t);

  procedure read32 (signal vme_in : out t_vme64x_in;
                    signal vme_out : in t_vme64x_out;
                    signal vme_timeout : boolean;
                    addr : std_logic_vector (31 downto 0);
                    am : vme_am_t;
                    variable data : out lword_t);

  procedure read64_2esst (signal vme_in : out t_vme64x_in;
                          signal vme_out : in t_vme64x_out;
                          signal vme_timeout : boolean;
                          addr : std_logic_vector (31 downto 0);
                          xam : vme_xam_t;
                          rate : std_logic_vector(3 downto 0);
                          variable data : out qword_array_t);

  procedure Dump_CR (signal vme_in : out t_vme64x_in;
                     signal vme_out : in t_vme64x_out;
                     signal vme_timeout : boolean);
end vme64x_sim_pkg;

library ieee;
use ieee.numeric_std.all;
use std.textio.all;

package body vme64x_sim_pkg is
  function hex1 (v : std_logic_vector (3 downto 0)) return character is
  begin
    case v is
      when x"0" => return '0';
      when x"1" => return '1';
      when x"2" => return '2';
      when x"3" => return '3';
      when x"4" => return '4';
      when x"5" => return '5';
      when x"6" => return '6';
      when x"7" => return '7';
      when x"8" => return '8';
      when x"9" => return '9';
      when x"a" => return 'a';
      when x"b" => return 'b';
      when x"c" => return 'c';
      when x"d" => return 'd';
      when x"e" => return 'e';
      when x"f" => return 'f';
      when "ZZZZ" => return 'Z';
      when "XXXX" => return 'X';
      when others => return '?';
    end case;
  end hex1;

  function hex2 (v : byte_t) return string is
  begin
    return hex1 (v(7 downto 4)) & hex1 (v(3 downto 0));
  end hex2;

  function hex6 (v : std_logic_vector (23 downto 0)) return string is
    variable res : string (6 downto 1);
  begin
    for i in res'range loop
      res (i) := hex1 (v (i * 4 - 1 downto i * 4 - 4));
    end loop;
    return res;
  end hex6;

  function hex8 (v : std_logic_vector (31 downto 0)) return string is
    variable res : string (8 downto 1);
  begin
    for i in res'range loop
      res (i) := hex1 (v (i * 4 - 1 downto i * 4 - 4));
    end loop;
    return res;
  end hex8;

  function hex (v : std_logic_vector) return string
  is
    constant ndigits : natural := v'length / 4;
    subtype av_t is std_logic_vector (ndigits * 4 - 1 downto 0);
    alias av : av_t is v;
    variable res : string (ndigits downto 1);
  begin
    for i in res'range loop
      res (i) := hex1 (av (i * 4 - 1 downto i * 4 - 4));
    end loop;
    return res;
  end hex;

  function Image_AM (am : natural range 0 to 63) return string is
  begin
    case am is
      when 16#3f# => return "A24S-BLT";
      when 16#3e# => return "A24S-PRG";
      when 16#3d# => return "A24S-DAT";
      when 16#3c# => return "A24S-MBL";
      when 16#3b# => return "A24U-BLT";
      when 16#3a# => return "A24U-PRG";
      when 16#39# => return "A24U-DAT";
      when 16#38# => return "A24U-MBL";

      when 16#37# => return "A40x-BLT";
      when 16#36# => return "Resvd-46";
      when 16#35# => return "A40x-LCK";
      when 16#34# => return "A40x-TFR";
      when 16#33# => return "Resvd-33";
      when 16#32# => return "A24x-LCK";
      when 16#31# => return "Resvd-31";
      when 16#30# => return "Resvd-30";

      when 16#2f# => return "CR-CSR  ";
      when 16#2e# => return "Resvd-2e";
      when 16#2d# => return "A16S    ";
      when 16#2c# => return "A16x-LCK";
      when 16#2b# => return "Resvd-2b";
      when 16#2a# => return "Resvd-2a";
      when 16#29# => return "A16U    ";
      when 16#28# => return "Resvd-28";

      when 16#27# downto 16#20# => return "Resvd-2x";
      when 16#1f# downto 16#10# => return "UD-1x";

      when 16#0f# => return "A32S-BLT";
      when 16#0e# => return "A32S-PRG";
      when 16#0d# => return "A32S-DAT";
      when 16#0c# => return "A32S-MBL";
      when 16#0b# => return "A32U-BLT";
      when 16#0a# => return "A32U-PRG";
      when 16#09# => return "A32U-DAT";
      when 16#08# => return "A32U-MBL";

      when 16#07# => return "Resvd-07";
      when 16#06# => return "Resvd-06";
      when 16#05# => return "A32x-LCK";
      when 16#04# => return "A64x-LCK";
      when 16#03# => return "A64x-BLT";
      when 16#02# => return "Resvd-02";
      when 16#01# => return "A64x-TFR";
      when 16#00# => return "A64x-MBL";
    end case;
  end Image_AM;

  --  Bus timer.  See VME spec 2.3.3 Bus Timer
  procedure bus_timer (signal vme_in : in  t_vme64x_in;
                       signal timeout : out boolean;
                       signal clk : std_logic)
  is
    type state_t is (IDLE, WAIT_DS, COUNTING, WAIT_END, ERR);
    variable state : state_t;
    variable count : natural;
  begin
    loop
    if rising_edge (clk) then
      if VME_in.RST_n = '0' then
        state := IDLE;
        timeout <= false;
      else
        timeout <= false;

        case state is
          when IDLE =>
            if VME_in.AS_n = '0' then
              state := WAIT_DS;
            end if;
          when WAIT_DS =>
            count := 20;
            if VME_in.DS_n /= "11" then
              state := COUNTING;
            end if;
          when COUNTING =>
            if VME_in.DS_n = "11" then
              state := WAIT_END;
            else
              if count = 0 then
                state := ERR;
              else
                count := count - 1;
              end if;
            end if;
          when WAIT_END =>
            if VME_in.AS_n = '1' then
              state := IDLE;
            end if;

          when ERR =>
            timeout <= true;
            -- VITAL 1 Rule 2.60
            -- Once it has driven BERR* low, the bus timer MUST NOT release
            -- BERR* until it detected both DS0* and DS1* high.
            if VME_in.DS_n = "11" then
              state := IDLE;
            end if;
        end case;
      end if;
    end if;
  end loop;
  end bus_timer;

  -- Convert a CR/CSR address to the VME address: insert GA.
  -- The ADDR is on 20 bits (so the x"" notation can be used), but as
  -- ADDR(19) is stipped, it must be '0'.
  function to_vme_cfg_addr (addr : cfg_addr_t)
    return std_logic_vector is
  begin
    assert addr (19) = '0' report "a19 is discarded" severity error;
    return x"00" & slave_ga & addr (18 downto 0);
  end to_vme_cfg_addr;

  procedure read_setup_addr (signal vme_in : out t_vme64x_in;
                             signal vme_out : in t_vme64x_out;
                             addr : std_logic_vector (31 downto 0);
                             am : vme_am_t) is
  begin
    VME_in.ADDR <= addr (31 downto 1);
    VME_in.AM <= am;
    VME_in.IACK_n <= '1';
    wait for 35 ns;
    VME_in.AS_n <= '0';
    VME_in.WRITE_n <= '1';
    if not (VME_out.DTACK_OE = '0' and VME_out.BERR_n = '1') then
      wait until VME_out.DTACK_OE = '0' and VME_out.BERR_n = '1';
    end if;
  end read_setup_addr;

  procedure read_wait_dtack (signal vme_out : in t_vme64x_out;
                             signal vme_timeout : boolean) is
  begin
    wait until (VME_out.DTACK_OE = '1' and VME_out.DTACK_n = '0')
      or vme_timeout;
    if not vme_timeout then
      assert VME_out.DATA_DIR = '1' report "bad data_dir";
      assert VME_out.DATA_OE_N = '0' report "bad data_oe_n";
    end if;
  end read_wait_dtack;

  procedure read_release (signal vme_in : out t_vme64x_in) is
  begin
    --  Release
    VME_in.AS_n <= '1';
    VME_in.DS_n <= "11";
  end read_release;

  procedure read8 (signal vme_in : out t_vme64x_in;
                   signal vme_out : in t_vme64x_out;
                   signal vme_timeout : boolean;
                   addr : std_logic_vector (31 downto 0);
                   am : vme_am_t;
                   variable data : out byte_t)
  is
    variable res : byte_t;
  begin
    if c_log then
      write (output, "read8 at 0x" & hex (addr) & " ["
             & Image_AM (to_integer (unsigned (am))) & " ]" & LF);
    end if;

    VME_in.LWORD_n <= '1';
    read_setup_addr (vme_in, vme_out, addr, am);

    if addr (0) = '0' then
      VME_in.DS_n <= "01";
    else
      VME_in.DS_n <= "10";
    end if;
    read_wait_dtack (vme_out, vme_timeout);
    if not vme_timeout then
      if addr (0) = '0' then
        res := VME_out.DATA (15 downto 8);
      else
        res := VME_out.DATA (7 downto 0);
      end if;
    else
      res := (others => 'X');
    end if;
    data := res;

    read_release (vme_in);

    if c_log then
      write (output, " => 0x" & hex(res) & LF);
    end if;
  end read8;

  procedure read16 (signal vme_in : out t_vme64x_in;
                    signal vme_out : in t_vme64x_out;
                    signal vme_timeout : boolean;
                    addr : std_logic_vector (31 downto 0);
                    am : vme_am_t;
                    variable data : out word_t)
    is
      variable res : word_t;
    begin
      if c_log then
        write (output, "read16 at 0x" & hex (addr) & " ["
               & Image_AM (to_integer (unsigned (am))) & " ]" & LF);
      end if;

      VME_in.LWORD_n <= '1';
      read_setup_addr (vme_in, vme_out, addr, am);

      assert addr (0) = '0' report "unaligned read16" severity error;
      VME_in.DS_n <= "00";
      read_wait_dtack (vme_out, vme_timeout);
      if not vme_timeout then
        res := VME_out.DATA (15 downto 0);
      else
        res := (others => 'X');
      end if;
      data := res;

      read_release (vme_in);

      if c_log then
        write (output, " => 0x" & hex(res) & LF);
      end if;
    end read16;

    procedure read32 (signal vme_in : out t_vme64x_in;
                      signal vme_out : in t_vme64x_out;
                      signal vme_timeout : boolean;
                      addr : std_logic_vector (31 downto 0);
                      am : vme_am_t;
                      variable data : out lword_t)
    is
      variable res : lword_t;
    begin
      if c_log then
        write (output, "read32 at 0x" & hex (addr) & " ["
               & Image_AM (to_integer (unsigned (am))) & " ]" & LF);
      end if;

      VME_in.LWORD_n <= '0';
      read_setup_addr (vme_in, vme_out, addr, am);

      assert addr (1 downto 0) = "00" report "unaligned read32" severity error;
      VME_in.DS_n <= "00";
      read_wait_dtack (vme_out, vme_timeout);
      if not vme_timeout then
        res := VME_out.DATA (31 downto 0);
      else
        res := (others => 'X');
      end if;
      data := res;

      read_release (vme_in);

      if c_log then
        write (output, " => 0x" & hex(res) & LF);
      end if;
    end read32;

    procedure read8_conf (signal vme_in : out t_vme64x_in;
                          signal vme_out : in t_vme64x_out;
                          signal vme_timeout : boolean;
                          addr : cfg_addr_t;
                          variable data : out byte_t) is
    begin
      read8 (vme_in, vme_out, vme_timeout,
             to_vme_cfg_addr (addr), c_AM_CR_CSR, data);
    end read8_conf;

    procedure read8_conf_mb (signal vme_in : out t_vme64x_in;
                             signal vme_out : in t_vme64x_out;
                             signal vme_timeout : boolean;
                             addr : cfg_addr_t;
                             variable data : out std_logic_vector)
    is
      variable ad : cfg_addr_t := addr;
      constant bsize : natural := data'length / 8;
      subtype d_type is std_logic_vector (bsize * 8 - 1 downto 0);
      alias d : d_type is data;
    begin
      assert data'length mod 8 = 0 report "data is not a multiple of bytes"
        severity error;
      for i in bsize - 1 downto 0 loop
        read8_conf (vme_in, vme_out, vme_timeout,
                    ad, d (i * 8 + 7 downto i * 8));
        ad := cfg_addr_t (unsigned(ad) + 4);
      end loop;
    end read8_conf_mb;

    procedure write_setup_addr (signal vme_in : out t_vme64x_in;
                                signal vme_out : in t_vme64x_out;
                                addr : std_logic_vector (31 downto 0);
                                lword_n : std_logic;
                                am : vme_am_t) is
    begin
      if not (VME_out.DTACK_OE = '0' and VME_out.BERR_n = '1') then
        wait until VME_out.DTACK_OE = '0' and VME_out.BERR_n = '1';
      end if;
      VME_in.ADDR <= addr(31 downto 1);
      VME_in.AM <= am;
      VME_in.LWORD_n <= lword_n;
      VME_in.IACK_n <= '1';
      wait for 35 ns;
      VME_in.AS_n <= '0';
      VME_in.WRITE_n <= '0';
    end write_setup_addr;

    procedure write_wait_dtack_pulse (signal vme_in : out t_vme64x_in;
                                      signal vme_out : in t_vme64x_out) is
    begin
      wait until VME_out.DTACK_OE = '1' and VME_out.DTACK_n = '0';
      VME_in.DS_n <= "11";

      wait until VME_out.DTACK_OE = '0' or VME_out.DTACK_n = '1';
    end write_wait_dtack_pulse;

    procedure write_wait_dtack (signal vme_in : out t_vme64x_in;
                                signal vme_out : in t_vme64x_out) is
    begin
      write_wait_dtack_pulse (vme_in, vme_out);

      VME_in.AS_n <= '1';
    end write_wait_dtack;

    procedure write8 (signal vme_in : out t_vme64x_in;
                      signal vme_out : in t_vme64x_out;
                      signal vme_timeout : boolean;
                      addr : std_logic_vector (31 downto 0);
                      am : vme_am_t;
                      data : byte_t) is
    begin
      write_setup_addr (vme_in, vme_out, addr, '1', am);

      if addr (0) = '0' then
        VME_in.DS_n <= "01";
        VME_in.DATA (15 downto 8) <= data;
      else
        VME_in.DS_n <= "10";
        VME_in.DATA (7 downto 0) <= data;
      end if;

      write_wait_dtack (vme_in, vme_out);
    end write8;

    procedure write16 (signal vme_in : out t_vme64x_in;
                       signal vme_out : in t_vme64x_out;
                       signal vme_timeout : boolean;
                       addr : std_logic_vector (31 downto 0);
                       am : vme_am_t;
                       data : word_t) is
    begin
      assert addr(0) = '0' report "unaligned write16" severity error;

      write_setup_addr (vme_in, vme_out, addr, '1', am);

      VME_in.DS_n <= "00";
      VME_in.DATA (15 downto 0) <= data;

      write_wait_dtack (vme_in, vme_out);
    end write16;

    procedure write32 (signal vme_in : out t_vme64x_in;
                       signal vme_out : in t_vme64x_out;
                       signal vme_timeout : boolean;
                       addr : std_logic_vector (31 downto 0);
                       am : vme_am_t;
                       data : lword_t) is
    begin
      assert addr(1 downto 0) = "00"
        report "unaligned write16" severity error;

      write_setup_addr (vme_in, vme_out, addr, '0', am);

      VME_in.DS_n <= "00";
      VME_in.DATA (31 downto 0) <= data;

      write_wait_dtack (vme_in, vme_out);
    end write32;

    procedure write8_conf (signal vme_in : out t_vme64x_in;
                           signal vme_out : in t_vme64x_out;
                           signal vme_timeout : boolean;
                           addr : cfg_addr_t;
                           data : byte_t) is
    begin
      if c_log then
        write (output,
               "write8_conf at 0x" & hex (addr) & " <= " & hex2 (data));
      end if;

      write8(vme_in, vme_out, vme_timeout,
             to_vme_cfg_addr (addr), c_AM_CR_CSR, data);
    end write8_conf;

    procedure read64_2esst (signal vme_in : out t_vme64x_in;
                            signal vme_out : in t_vme64x_out;
                            signal vme_timeout : boolean;
                            addr : std_logic_vector (31 downto 0);
                            xam : vme_xam_t;
                            rate : std_logic_vector(3 downto 0);
                            variable data : out qword_array_t)
    is
      variable odd : boolean;
    begin
      assert addr (3 downto 0) = "0000"
        report "unaligned read64_2e" severity error;
      assert data'length mod 2 = 0 severity error;

      --  Phase 1.
      VME_in.LWORD_n <= xam(0);
      read_setup_addr (vme_in, vme_out,
                       addr (31 downto 8) & xam (7 downto 1) & '0', b"10_0000");
      VME_in.DATA <= x"0000_0000";
      VME_in.DS_n <= "10";
      wait until (VME_out.DTACK_OE = '1' and VME_out.DTACK_n = '0')
        or vme_timeout;
      --  Phase 2.
      VME_in.ADDR <= x"01" & x"00" & std_logic_vector(to_unsigned(data'length / 2, 8)) & addr (7 downto 1);
      VME_in.DATA <= x"00_00_00_0" & rate;
      VME_in.LWORD_n <= addr (0);
      wait for 5 ns;
      VME_in.DS_n <= "11";
      wait until (VME_out.DTACK_OE = '1' and VME_out.DTACK_n = '1')
        or vme_timeout;
      --  Phase 3.
      VME_in.ADDR <= (others => '0');
      VME_in.DATA <= x"0000_0000";
      wait for 5 ns;
      VME_in.DS_n <= "10";
      wait until (VME_out.DTACK_OE = '1' and VME_out.DTACK_n = '0')
        or vme_timeout;

      VME_in.DS_n <= "00";

      odd := true;
      for i in data'range loop
        if odd then
          wait until (VME_out.DTACK_OE = '1' and VME_out.DTACK_n = '1')
            or vme_timeout;
        else
          wait until (VME_out.DTACK_OE = '1' and VME_out.DTACK_n = '0')
            or vme_timeout;
        end if;
        odd := not odd;
        data (i) := VME_out.ADDR & VME_out.LWORD_n & VME_out.DATA;
      end loop;

      --  2eSST 2.8 Transfer termination
      --  Transfers always terminate on an even beat.
      --
      --  The slave terminates a read transfer by stopping the
      --  data transmission and then asserting RETRY*/RESP* and
      --  BERR*.
      wait until (VME_out.BERR_n = '0'
                  and VME_out.RETRY_n = '0' and VME_out.RETRY_OE = '1');

      --  The master will respond by negating DS0*
      VME_in.DS_n <= "11";

      read_release (vme_in);
    end read64_2esst;

    procedure ack_int (signal vme_in : out t_vme64x_in;
                       signal vme_out : in t_vme64x_out;
                       signal vme_timeout : boolean;
                       vec : out byte_t) is
    begin
      if VME_out.IRQ_n = "1111111" then
        vec := (others => 'X');
        return;
      end if;

      for i in 6 downto 0 loop
        if VME_out.IRQ_n (i) = '0' then
          VME_in.ADDR (3 downto 1) <= std_logic_vector (to_unsigned (i + 1, 3));
          exit;
        end if;
      end loop;
      VME_in.IACK_n <= '0';
      VME_in.WRITE_n <= '1';
      wait for 35 ns;
      VME_in.AS_n <= '0';
      if not (VME_out.DTACK_OE = '0' and VME_out.BERR_n = '1') then
        wait until VME_out.DTACK_OE = '0' and VME_out.BERR_n = '1';
      end if;

      VME_in.DS_n <= "10";
      read_wait_dtack (vme_out, vme_timeout);
      if not vme_timeout then
        vec := VME_out.DATA (7 downto 0);
      else
        vec := (others => 'X');
      end if;

      read_release (vme_in);
      VME_in.IACK_n <= '1';
    end ack_int;

    function Disp_DAWPR (v : byte_t) return string is
  begin
    case v is
      when x"81" => return "D08(O)";
      when x"82" => return "D08(EO)";
      when x"83" => return "D16 + D08";
      when x"84" => return "D32 + D16 + D08";
      when x"85" => return "MD32 + D16 + D08";
      when others => return "??";
    end case;
  end Disp_DAWPR;

    procedure Disp_AMCAP (cap : std_logic_vector (63 downto 0)) is
    begin
      if cap = (cap'range => '0') then
        write (output, " -");
      else
        for i in cap'range loop
          if cap (i) = '1' then
            write (output, " ");
            write (output, Image_AM (i));
          end if;
        end loop;
      end if;
    end Disp_AMCAP;

    procedure Dump_CR (signal vme_in : out t_vme64x_in;
                       signal vme_out : in t_vme64x_out;
                       signal vme_timeout : boolean)
    is
      variable d : byte_t;
      variable b3 : std_logic_vector (23 downto 0);
      variable w : std_logic_vector (31 downto 0);
      variable b8 : std_logic_vector (63 downto 0);
      variable b32 : std_logic_vector (255 downto 0);
    begin
      if false then
      read8_conf (vme_in, vme_out, vme_timeout, x"0_0003", d);
      write (output, "CR checksum:     0x" & hex2 (d) & LF);

      read8_conf (vme_in, vme_out, vme_timeout, x"0_0013", d);
      assert d = x"81" report "invalid data width at 0x13" severity error;
      write (output, "CR data width:   0x" & hex2 (d) & LF);

      read8_conf (vme_in, vme_out, vme_timeout, x"0_0017", d);
      assert d = x"81" report "invalid data width at 0x17" severity error;
      write (output, "CSR data width:  0x" & hex2 (d) & LF);

      read8_conf (vme_in, vme_out, vme_timeout, x"0_001b", d);
      assert d = x"02" report "invalid spec ID at 0x1b" severity error;
      write (output, "CR/CSR version:  0x" & hex2 (d) & LF);

      read8_conf (vme_in, vme_out, vme_timeout, x"0_001f", d);
      assert d = x"43" report "invalid valid CR byte at 0x1f" severity error;
      write (output, "CR valid 'C':    0x" & hex2 (d) & LF);

      read8_conf (vme_in, vme_out, vme_timeout, x"0_0023", d);
      assert d = x"52" report "invalid valid CR byte at 0x23" severity error;
      write (output, "CR valid 'R':    0x" & hex2 (d) & LF);

      read8_conf_mb (vme_in, vme_out, vme_timeout, x"0_0027", b3);
      write (output, "CR Manu ID:      0x" & hex6 (b3) & LF);

      read8_conf_mb (vme_in, vme_out, vme_timeout, x"0_0033", w);
      write (output, "CR Board ID:     0x" & hex8 (w) & LF);

      read8_conf_mb (vme_in, vme_out, vme_timeout, x"0_0043", w);
      write (output, "CR Rev ID:       0x" & hex8 (w) & LF);

      read8_conf_mb (vme_in, vme_out, vme_timeout, x"0_0053", b3);
      write (output, "CR ASCII ptr:    0x" & hex6 (b3) & LF);

      read8_conf (vme_in, vme_out, vme_timeout, x"0_007F", d);
      write (output, "CR Program ID:   0x" & hex2 (d) & LF);
      end if;

      --  VME64x
      if false then
      read8_conf_mb (vme_in, vme_out, vme_timeout, x"0_0083", b3);
      write (output, "CR BEG_USER_CR:  0x" & hex6 (b3) & LF);

      read8_conf_mb (vme_in, vme_out, vme_timeout, x"0_008F", b3);
      write (output, "CR END_USER_CR:  0x" & hex6 (b3) & LF);

      read8_conf_mb (vme_in, vme_out, vme_timeout, x"0_009B", b3);
      write (output, "CR BEG_CRAM:     0x" & hex6 (b3) & LF);

      read8_conf_mb (vme_in, vme_out, vme_timeout, x"0_00A7", b3);
      write (output, "CR END_CRAM:     0x" & hex6 (b3) & LF);

      read8_conf_mb (vme_in, vme_out, vme_timeout, x"0_00B3", b3);
      write (output, "CR BEG_USER_CSR: 0x" & hex6 (b3) & LF);

      read8_conf_mb (vme_in, vme_out, vme_timeout, x"0_00B3", b3);
      write (output, "CR END_USER_CSR: 0x" & hex6 (b3) & LF);

      read8_conf_mb (vme_in, vme_out, vme_timeout, x"0_00CB", b3);
      write (output, "CR BEG_SN:       0x" & hex6 (b3) & LF);

      read8_conf_mb (vme_in, vme_out, vme_timeout, x"0_00D7", b3);
      write (output, "CR END_SN:       0x" & hex6 (b3) & LF);

      read8_conf (vme_in, vme_out, vme_timeout, x"0_00E3", d);
      write (output, "CR Slave charac: 0x" & hex2 (d) & LF);

      read8_conf (vme_in, vme_out, vme_timeout, x"0_00E7", d);
      write (output, "CR user-def:     0x" & hex2 (d) & LF);

      read8_conf (vme_in, vme_out, vme_timeout, x"0_00EB", d);
      write (output, "CR master chara: 0x" & hex2 (d) & LF);

      read8_conf (vme_in, vme_out, vme_timeout, x"0_00EF", d);
      write (output, "CR user-def:     0x" & hex2 (d) & LF);

      read8_conf (vme_in, vme_out, vme_timeout, x"0_00F3", d);
      write (output, "CR INT hand cap: 0x" & hex2 (d) & LF);

      read8_conf (vme_in, vme_out, vme_timeout, x"0_00F7", d);
      write (output, "CR INT cap:      0x" & hex2 (d) & LF);

      read8_conf (vme_in, vme_out, vme_timeout, x"0_00FF", d);
      write (output, "CR CRAM acc wd:  0x" & hex2 (d) & LF);

      for i in 0 to 7 loop
        read8_conf (vme_in, vme_out, vme_timeout,
                    std_logic_vector (unsigned'(x"0_0103") + i * 4), d);
        write (output, "CR FN" & natural'image (i) & " DAW:     0x"
               & hex2 (d) & "  [" & Disp_DAWPR (d) & ']' & LF);
      end loop;

      for i in 0 to 7 loop
        read8_conf_mb (vme_in, vme_out, vme_timeout,
                       std_logic_vector (unsigned'(x"0_0123") + i * 32), b8);
        write (output, "CR FN" & natural'image (i) & " AMCAP:    0x"
               & hex (b8) & " [");
        Disp_AMCAP (b8);
        write (output, " ]" & LF);
      end loop;
      end if;

      read8_conf_mb (vme_in, vme_out, vme_timeout, x"0_0223", b32);
      write (output, "CR FN0 XAMCAP:   0x" & hex (b32) & LF);
      read8_conf_mb (vme_in, vme_out, vme_timeout, x"0_02a3", b32);
      write (output, "CR FN1 XAMCAP:   0x" & hex (b32) & LF);
      --  ...

      read8_conf_mb (vme_in, vme_out, vme_timeout, x"0_0623", w);
      write (output, "CR FN0 ADEM:     0x" & hex (w) & LF);
      read8_conf_mb (vme_in, vme_out, vme_timeout, x"0_0633", w);
      write (output, "CR FN1 ADEM:     0x" & hex (w) & LF);
      -- ...

      read8_conf_mb (vme_in, vme_out, vme_timeout, x"7_FF63", w);
      write (output, "CR FN0 ADER:     0x" & hex (w) & LF);
      read8_conf_mb (vme_in, vme_out, vme_timeout, x"7_FF73", w);
      write (output, "CR FN1 ADER:     0x" & hex (w) & LF);

      read8_conf (vme_in, vme_out, vme_timeout, x"0_06af", d);
      write (output, "CR master DAWPR: 0x" & hex2 (d) & LF);

      read8_conf_mb (vme_in, vme_out, vme_timeout, x"0_06b3", b8);
      write (output, "CR master AMCAP: 0x" & hex (b8) & LF);

      read8_conf_mb (vme_in, vme_out, vme_timeout, x"0_06d3", b32);
      write (output, "CR mastr XAMCAP: 0x" & hex (b32) & LF);
    end Dump_CR;
end vme64x_sim_pkg;
