#!/usr/bin/env python
import PyUAL
import getopt, sys

AM_map = {0x3f: "A24S-BLT",
          0x3e: "A24S-PRG",
          0x3d: "A24S-DAT",
          0x3c: "A24S-MBL",
          0x3b: "A24U-BLT",
          0x3a: "A24U-PRG",
          0x39: "A24U-DAT",
          0x38: "A24U-MBL",

          0x37: "A40-BLT ",
          0x36: "Resvd-36",
          0x35: "A40-LCK ",
          0x34: "A40     ",
          0x33: "Resvd-33",
          0x32: "A24-LCK ",
          0x31: "Resvd-31",
          0x30: "Resvd-30",

          0x2f: "CR/CSR  ",
          0x2e: "Resvd-2e",
          0x2d: "A16S    ",
          0x2c: "A16-LCK ",
          0x2b: "Resvd-2b",
          0x2a: "Resvd-2a",
          0x29: "A16U    ",
          0x28: "Resvd-28",

          0x27: "Resvd-27",
          0x26: "Resvd-26",
          0x25: "Resvd-25",
          0x24: "Resvd-24",
          0x23: "Resvd-23",
          0x22: "Resvd-22",
          0x21: "Resvd-21",
          0x20: "Resvd-20",

          0x1f: "Userd-1f",
          0x1e: "Userd-1e",
          0x1d: "Userd-1d",
          0x1c: "Userd-1c",
          0x1b: "Userd-1b",
          0x1a: "Userd-1a",
          0x19: "Userd-19",
          0x18: "Userd-18",

          0x17: "Userd-17",
          0x16: "Userd-16",
          0x15: "Userd-15",
          0x14: "Userd-14",
          0x13: "Userd-13",
          0x12: "Userd-12",
          0x11: "Userd-11",
          0x10: "Userd-10",

          0x0f: "A32S-BLT",
          0x0e: "A32S-PRG",
          0x0d: "A32S-DAT",
          0x0c: "A32S-MBL",
          0x0b: "A32U-BLT",
          0x0a: "A32U-PRG",
          0x09: "A32U-DAT",
          0x08: "A32U-MBL",

          0x07: "Resvd-07",
          0x06: "Resvd-06",
          0x05: "A32-LCK ",
          0x04: "A64-LCK ",
          0x03: "A64-BLT ",
          0x02: "Resvd-02",
          0x01: "A64     ",
          0x00: "A64-MBLT"}

DAWPR_map = { 0x00: "not implemented",
              0x81: "D08(O)",
              0x82: "D08(EO)",
              0x83: "D16 or D08(EO)",
              0x84: "D32, D16 or D08(EO)",
              0x85: "MD32, D16 or D08(EO)",
              0xff: "not to be used" }

def get_DAWPR_name(v):
    return DAWPR_map.get(v, "reserved")


def get_CDAW_name(v):
    CDAW_map = { 0x00: "not to be used",
                 0x81: "D08(EO), every fourth byte",
                 0x82: "D08(EO), every other byte",
                 0x83: "D16 or D08(EO), every byte",
                 0x84: "D32, D16, D08(EO), every byte",
                 0xff: "not to be used" }
    return CDAW_map.get(v, "reserved")

def get_program_id_name(v):
    id_map = { 0x00: "not used",
               0x01: "IDROM only",
               0xff: "not to be used" }
    return id_map.get(v, "-")

def get_spec_id_name(v):
    id_map = { 0x00: "not to be used",
               0x01: "VME64-1994",
               0x02: "VME64x",
               0xff: "not to be used" }
    return id_map.get(v, "reserved")

def read_mb(cr_csr, offset, length):
    res = 0
    for i in range(length):
        res *= 0x100
        res |= cr_csr.readb(offset)
        offset += 4
    return res

def read_adem(cr_csr, num):
    assert num >= 0 and num <= 7
    adem = read_mb(cr_csr, 0x623 + num * 0x10, 4)
    return {'raw': adem,
            'addr': adem & 0xffffff00,
            'faf': (adem >> 3) & 1,
            'dfs': (adem >> 2) & 1,
            'efd': (adem >> 1) & 1,
            'efm': (adem >> 0) & 1}

def read_ader(cr_csr, num):
    assert num >= 0 and num <= 7
    ader = read_mb(cr_csr, 0x7ff63 + num * 0x10, 4)
    return {'raw': ader,
            'addr': ader & 0xffffff00,
            'am': (ader >> 2) & 0x3f,
            'dfsr': (ader >> 1) & 1,
            'xam': ader & 1}

def disp_csr(cr_csr):
    print ("CSR:")
    print (" BAR        0x{0:02x}".format(cr_csr.readb(0x7ffff)))
    print (" Bit Reg    0x{0:02x}".format(cr_csr.readb(0x7fffb)))
    print (" CRAM Owner 0x{0:02x}".format(cr_csr.readb(0x7fff3)))
    print (" USR Reg    0x{0:02x}".format(cr_csr.readb(0x7ffef)))
    for i in range(8):
        ader = read_ader(cr_csr, i)
        print (" Func {0} ADER  0x{1:08x}  AM:{2:02x} {3}".format(
                i, ader['raw'], ader['am'], AM_map[ader['am']]))

def compute_checksum(cr_csr):
    rom_length = read_mb(cr_csr, 0x07, 3)
    # Compute the checksum: every fourth byte from 0x7 for rom_length bytes.
    # Not sure it is correct for CR DAW /= D08(O), but I have no such card to
    # test.
    chksum = 0;
    for i in range(rom_length):
        off = 0x07 + i * 4
        b = cr_csr.readb(off)
        chksum += b

    chksum &= 0xff
    return chksum

def validate_cr(cr_csr):
    if cr_csr.readb(0x1f) != 0x43:
        return False
    if cr_csr.readb(0x23) != 0x52:
        return False
    return True

def disp_cr(cr_csr):
    print ("CR:")
    if False:
        for i in range(0, 0x40, 4):
            print (" {0:04x}: {1:08x}".format(i, cr_csr.readl(i)))

    if not validate_cr(cr_csr):
        print (" Incorrect Config ROM")
        return

    rom_length = read_mb(cr_csr, 0x07, 3)
    chksum = compute_checksum(cr_csr)
    print (" checksum      0x{0:02x} vs 0x{1:02x}".format(
            cr_csr.readb(0x03), chksum))
    print (" ROM length    0x{0:06x}".format(rom_length))
    cr_daw = cr_csr.readb(0x13)
    print (" CR DAW        0x{0:02x}: {1}".format(
            cr_daw, get_CDAW_name(cr_daw)))
    csr_daw = cr_csr.readb(0x17)
    print (" CSR DAW       0x{0:02x}: {1}".format(
            csr_daw, get_CDAW_name(csr_daw)))
    spec_id = cr_csr.readb(0x1b)
    print (" Spec ID       0x{0:02x}: {1}".format(
            spec_id, get_spec_id_name(spec_id)))
    print (" C (0x43)      0x{0:02x}".format(cr_csr.readb(0x1f)))
    print (" R (0x52)      0x{0:02x}".format(cr_csr.readb(0x23)))
    manu_id = read_mb(cr_csr, 0x27, 3);
    print (" Manufacturer  0x{0:06x}".format(manu_id))
    board_id = read_mb(cr_csr, 0x33, 4)
    print (" Board ID      0x{0:08x}".format(board_id))
    print (" Revision ID   0x{0:08x}".format(read_mb(cr_csr, 0x43, 4)))
    str_off = read_mb(cr_csr, 0x53, 3)
    if str_off != 0:
        str = ""
        off = str_off
        while True:
            c = cr_csr.readb(off)
            if c == 0:
                break
            str += chr(c)
            off += 4
        print (" String off    0x{0:06x} - 0x{1:06x}".format(str_off, off))
        print ("   " + str)
    else:
        print (" String off    0x{0:06x}".format(str_off))

    prg_id = cr_csr.readb(0x7f)
    print (" Program ID    0x{0:02x}: {1}".format(
            prg_id, get_program_id_name(prg_id)))
    print (" BEG_USER_CR   0x{0:06x}".format(read_mb(cr_csr, 0x83, 3)))
    print (" END_USER_CR   0x{0:06x}".format(read_mb(cr_csr, 0x8f, 3)))
    print (" BEG_CRAM      0x{0:06x}".format(read_mb(cr_csr, 0x9b, 3)))
    print (" END_CRAM      0x{0:06x}".format(read_mb(cr_csr, 0xa7, 3)))
    beg_user_csr = read_mb(cr_csr, 0xb3, 3)
    print (" BEG_USER_CSR  0x{0:06x}".format(beg_user_csr))
    end_user_csr = read_mb(cr_csr, 0xbf, 3)
    print (" END_USER_CSR  0x{0:06x}".format(end_user_csr))
    print (" BEG_SN        0x{0:06x}".format(read_mb(cr_csr, 0xcb, 3)))
    print (" END_SN        0x{0:06x}".format(read_mb(cr_csr, 0xd7, 3)))
    print (" Slave Charac  0x{0:02x}".format(cr_csr.readb(0xe3)))
    print (" Usr Slave Ch  0x{0:02x}".format(cr_csr.readb(0xe7)))
    print (" Master Charac 0x{0:02x}".format(cr_csr.readb(0xeb)))
    print (" Usr Master Ch 0x{0:02x}".format(cr_csr.readb(0xef)))
    print (" INT hndl Ch   0x{0:02x}".format(cr_csr.readb(0xf3)))
    print (" INT Cap       0x{0:02x}".format(cr_csr.readb(0xf7)))
    CRAM_AW = cr_csr.readb(0xff)
    print (" CRAM Acc Wdth 0x{0:02x}: {1}".format(
            CRAM_AW, get_CDAW_name(CRAM_AW)))
    for i in range(8):
        daw = cr_csr.readb(0x103 + 4 * i)
        print (" Func {0} DAWPR  0x{1:02x}: {2}".format(
                i, daw, get_DAWPR_name(daw)))
    for i in range(8):
        am = read_mb(cr_csr, 0x123 + i * 0x20, 8)
        print (" Func {0} AM     {1:016x}".format(i, am))
        if am != 0:
            print "    ",
            for j in range(64):
                if (am >> j) & 1:
                    print AM_map[j],
            print
    for i in range(8):
        print (" Func {0} XAM    {1:064x}".format(
                i, read_mb(cr_csr, 0x223 + i * 0x80, 32)))
    for i in range(8):
        adem = read_adem(cr_csr, i)
        print (" Func {0} ADEM   0x{1:08x}".format(
                i, adem['raw']))
    print (" Master DAWPR  0x{0:02x}".format(cr_csr.readb(0x6af)))
    print (" Master AM     {0:016x}".format(read_mb(cr_csr, 0x6b3, 8)))
    print (" Master XAM    {0:064x}".format(read_mb(cr_csr, 0x6d3, 32)))

    if manu_id == 0x80030 and board_id == 0x198:
        print ("SVEC:")
        if beg_user_csr != 0 and end_user_csr - beg_user_csr >= 0x2c:
            print ("  IRQ level:  {0:02x}".format(
                    cr_csr.readb(beg_user_csr + 0x28)))
            print ("  IRQ vector: {0:02x}".format(
                    cr_csr.readb(beg_user_csr + 0x2c)))

def main():
    slot = None
    try:
        opts, args = getopt.getopt(sys.argv[1:], "s:")
    except getopt.GetoptError as err:
        print str(err)
        sys.exit(2)
    for o, a in opts:
        if o == "-s":
            slot = int(a)
        else:
            assert False, "unhandled option"
    
    if slot:
        cr_csr_desc = PyUAL.PyUALVME(4, 0x2f, 0x80000, slot << 19, 0)
        #print (cr_csr_desc)
        cr_csr = PyUAL.PyUAL("vme", cr_csr_desc)
        disp_cr(cr_csr)
        disp_csr(cr_csr)
    else:
        for i in range(1,32):
            cr_csr_desc = PyUAL.PyUALVME(4, 0x2f, 0x80000, i << 19, 0)
            cr_csr = PyUAL.PyUAL("vme", cr_csr_desc)

            bar = cr_csr.readb(0x7ffff)
            if bar == 0xff:
                continue
            print ("CSR {0:2d} BAR: 0x{1:02x}".format(i, bar))

if __name__ == "__main__":
    main()

